!! Copyright (C) 2004-2010 M. Oliveira, F. Nogueira
!!
!! This program is free software; you can redistribute it and/or modify
!! it under the terms of the GNU General Public License as published by
!! the Free Software Foundation; either version 2, or (at your option)
!! any later version.
!!
!! This program is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU General Public License for more details.
!!
!! You should have received a copy of the GNU General Public License
!! along with this program; if not, write to the Free Software
!! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
!! 02111-1307, USA.

#include "global.h"

module wave_equations_integrator
  use global
  use messages
  use io
  use gsl_interface
  use oct_parser
  use utilities
  use quantum_numbers
  use potentials
  use wave_equations_derivs
  implicit none


                    !---Interfaces---!

  interface
    function ode_solve(x2, stp, evl, ctrl, h1, hmin, nstepmax, nstep, x, y1, y2, myfunc)
      use global
      implicit none
      integer(POINTER_SIZE), intent(inout) :: stp, evl, ctrl
      integer,               intent(in)    :: nstepmax
      integer,               intent(out)   :: nstep
      real(R8),              intent(in)    :: x2, h1, hmin
      real(R8),              intent(inout) :: x(nstepmax), y1(nstepmax), y2(nstepmax)
      integer :: ode_solve
      interface
        subroutine myfunc(t, y, f)
          use global
          real(R8), intent(in)  :: t
          real(R8), intent(in)  :: y(2)
          real(R8), intent(out) :: f(2)
        end subroutine myfunc
      end interface
    end function ode_solve
  end interface


                    !---Derived Data Types---!

  type integrator_type
    private
    integer               :: nstepmax
    integer(POINTER_SIZE) :: evl, stp, ctrl
    real(R8)              :: hmin, tol
  end type integrator_type


                    !---Global Variables---!

  integer, parameter :: SCHRODINGER = EQ1, &
                        DIRAC       = EQ2, &
                        SCALAR_REL  = EQ3

  integer, parameter :: M_RK2   = 1, &
                        M_RK4   = 2, &
                        M_RKF4  = 3, &
                        M_RKCK4 = 4, &
                        M_RKPD8 = 5


  !Wavefunctions values at infinity
  real(R8), parameter :: SCHRODINGER_FINF = 1.0E-20_r8
  real(R8), parameter :: DIRAC_FINF       = 1.0E-20_r8


                    !---Public/Private Statements---!

  private
  public :: integrator_type, &
            wave_equations_integrator_init, &
            outward_integration, &
            inward_integration, &
            practical_infinity, &
            wave_equations_integrator_end, &
            SCHRODINGER, &
            DIRAC, &
            SCALAR_REL, &
            SCHRODINGER_FINF, &
            DIRAC_FINF, &
            M_RK2, &
            M_RK4, &
            M_RKF4, &
            M_RKCK4, &
            M_RKPD8

contains

  subroutine wave_equations_integrator_init(integrator_sp, integrator_dp)
    !-----------------------------------------------------------------------!
    ! Initializes integrator objects by reading some parameters from the    !
    ! input file and by initializing GSL ODE solver objects.                !
    ! Two objects are initialized: one for single precision integrations    !
    ! and a second one for double precision integrations.                   !
    !                                                                       !
    !  integrator_sp - single-precision integrator object                   !
    !  integrator_dp - double-precision integrator object                   !
    !-----------------------------------------------------------------------!
    type(integrator_type), intent(out) :: integrator_sp, integrator_dp

    integer  :: stepping_func
    real(R8) :: tol

    call push_sub("wave_equations_integrator_init")

    message(1) = ""
    message(2) = "Initializing Wave-Equations Integrator"
    call write_info(2)

    !Which stepping function will we use
    call oct_parse_int('SteppingFunction', 5, stepping_func)
    select case (stepping_func)
    case (M_RK2)
      message(1) = "  Stepping function: Embedded 2nd order Runge-Kutta"
      message(2) = "                     with 3rd order error estimate"
      call write_info(2,20)
    case (M_RK4)
      message(1) = "  Stepping function: 4th order (classical) Runge-Kutta method"
      call write_info(1,20)
    case (M_RKF4)
      message(1) = "  Stepping function: Embedded 4th order Runge-Kutta-Fehlberg "
      message(2) = "                     method with 5th order error estimate"
      call write_info(2,20)
    case (M_RKCK4)
      message(1) = "  Stepping function: Embedded 4th order Runge-Kutta Cash-Karp"
      message(2) = "                     method with 5th order error estimate"
      call write_info(2,20)
    case (M_RKPD8)
      message(1) = "  Stepping function: Embedded 8th order Runge-Kutta Prince-Dormand"
      message(2) = "                     method with 9th order error estimate"
      call write_info(2,20)
!    case (6)
!      message(1) = "  Stepping function: Implicit 4th order Runge-Kutta at Gaussian points"
!      call write_info(1,20)
!    case (7)
!      message(1) = "  Stepping function: M=2 implicit Gear method"
!      call write_info(1,20)
    case default
       message(1) = "SteppingFunction must be between 1 and 5."
       call write_fatal(1)
    end select

    !Read the ODE integrator tolerance
    call oct_parse_double('ODEIntTolerance', 1.0E-6_r8, tol)
    if (tol <= M_ZERO) then
      message(1) = "ODEIntTolerance must be greater than zero."
      call write_fatal(1)
    end if
    write(message(1),'(2X,"ODE Integrator tolerance: ",ES10.3E2)') tol
    call write_info(1,20)

    !Read the maximum number of steps the ODE integrator is allowed to take
    call oct_parse_int('MaxSteps', 50000, integrator_sp%nstepmax)
    if (integrator_sp%nstepmax <= M_ZERO) then
      message(1) = "MaxSteps must be greater than zero."
      call write_fatal(1)
    end if
    write(message(1),'(2X,"ODE Integrator maximum number of steps: ",I6)') integrator_sp%nstepmax
    call write_info(1,20)

    !What is the minimum step size allowed
    call oct_parse_double('MinimumStep', 1.0E-15_r8, integrator_sp%hmin)
    if (integrator_sp%hmin <= M_ZERO) then
       message(1) = "MinimumStep must be greater than zero."
       call write_fatal(1)
    end if
    write(message(1),'(2X,"ODE Integrator minimum step size: ",ES10.3E2)') integrator_sp%hmin
    call write_info(1,20)

    integrator_dp = integrator_sp
    integrator_sp%tol = sqrt(tol)
    integrator_dp%tol = tol**2

    !Initialize GSL objects
    call gsl_odeiv_step_alloc(stepping_func, 2, integrator_sp%stp)
    call gsl_odeiv_evolve_alloc(2, integrator_sp%evl)
    call gsl_odeiv_control_standart_new(integrator_sp%ctrl, 1.0E-22_r8, integrator_sp%tol, M_ONE, M_ONE)

    call gsl_odeiv_step_alloc(stepping_func, 2, integrator_dp%stp)
    call gsl_odeiv_evolve_alloc(2, integrator_dp%evl)
    call gsl_odeiv_control_standart_new(integrator_dp%ctrl, 1.0E-22_r8, integrator_dp%tol, M_ONE, M_ONE)

    call pop_sub()
  end subroutine wave_equations_integrator_init

  subroutine outward_integration(qn, e, wave_eq, potential, integrator, nstep, r_out, g_out, f_out, rc)
    !-----------------------------------------------------------------------!
    ! Integrates the radial wave-equation outward from a point near the     !
    ! origin to the classical turning point.                                !
    !                                                                       !
    !  qn         - set of quantum numbers                                  !
    !  e          - energy                                                  !
    !  wave_eq    - wave-equation to integrate                              !
    !  potential  - potential object                                        !
    !  integrator - integrator object                                       !
    !  nstep      - number of steps taken by the solver                     !
    !  r_out      - mesh used by the ODE solver                             !
    !  g_out      - wavefunction                                            !
    !  f_out      - wavefunction                                            !
    !  rc         - if present the integration will stop at rc instead of   !
    !              stopping at the classical turning point                  !
    !-----------------------------------------------------------------------!
    type(qn_type),         intent(in)    :: qn
    real(R8),              intent(in)    :: e
    integer,               intent(in)    :: wave_eq
    type(potential_type),  intent(in)    :: potential
    type(integrator_type), intent(inout) :: integrator
    integer,               intent(out)   :: nstep
    real(R8),              pointer       :: r_out(:), g_out(:), f_out(:)
    real(R8),              intent(in), optional :: rc

    integer  :: alpha, beta, ierr
    real(R8) :: z, vp0, r0, ri, s, a0, a1, a2, a3, b0, b1, b2, w
    real(R8), allocatable :: r(:), g(:), f(:)

    call push_sub("outward_integration")

    !Allocate work arrays
    allocate(r(integrator%nstepmax), g(integrator%nstepmax), f(integrator%nstepmax))

    !Get the initial and final points
    r0 = smallest_safe_r(potential)
    if (present(rc)) then
      ri = rc
    else
      ri = classical_turning_point(potential, e, qn)
    end if

    ! For a potential v(r) = vp(r) - Z/r, the general solutions of the equations 
    ! when r->0 are of the form:
    !
    !   g(r) = r**(s-1) (a0 + a1 r + a2 r**2 + ...)
    !   f(r) = r**(s-1) (b0 + b1 r + b2 r**2 + ...)
    ! 
    ! Schrodinger equation:
    !   s = l
    !   a0 = 0
    !   a2 = - Z/(l + 1) a1
    !   a3 = [-Z a2 + (vp - e) a1]/(2l + 3)
    !   b0 = l a1
    !   b1 = (l + 1) a2
    !   b2 = (l + 2) a3
    !
    ! Scalar-relativistic equation:
    !   s  = sqrt(l(l + 1) + 1 - (Z/c)**2)
    !   a1 = [(e + 2c**2 - vp) b0 - Z/c (2e + 2c**2 - vp) a0]/[(2s + 1)c]
    !   a2 = [(e + 2c**2 - vp)(vp - e)/c a0 + (e + 2c**2 - vp) b1 - Z/c (2e + 2c**2 - vp) a1]/[2(2s + 2)c]
    !   b0 = c/Z (s - 1) a0
    !   b1 = c/Z [s a1 - (e + 2c**2 - vp)/c b0]
    !   b2 = c/Z [(s + 1) a2 - (e + 2c**2 - vp)/c b1]
    !
    ! Dirac equation:
    !   s  = sqrt(k**2 - (Z/c)**2)
    !   a0 = Z/c b0/(s + k)  v  a0 = - c/Z (s - k) b0
    !   a1 = [(s + 1 - k)(e + 2c**2 - vp) b0 - Z/c (e - vp) a0]/(2s + 1)/c
    !   a2 = [(s + 2 - k)(e + 2c**2 - vp) b1 - Z/c (e - vp) a1]/(2s + 2)/(2c)
    !   b0 = -Z/c a0/(s - k)  v  b0 = c/Z (s + k) a0
    !   b1 = [-(e - vp)/c a0 - Z/c a1]/(s + 1 - k)
    !   b2 = [-(e - vp)/c a1 - Z/c a2]/(s + 2 - k)
    ! 
    ! NOTE: for the scalar-relativistic equation and for Dirac's
    !       equation Z=0 is a special case.
    r(1) = r0
    z = potential_nuclear_charge(potential)    
    vp0 = v(potential, r0, qn) + z/r0
    select case (wave_eq)
    case (SCHRODINGER)
      s = qn%l
      a0 = M_ZERO
      a1 = M_ONE
      a2 = -z/(s + M_ONE)*a1
      a3 = (-z*a2 + (vp0 - e)*a1)/(M_TWO*s + M_THREE)
      b0 = s*a1
      b1 = (s + 1)*a2
      b2 = (s + 2)*a3
    case (DIRAC)
      if (z /= M_ZERO) then
        a0 = M_ONE
        s = sqrt(qn%k**2 - (z/M_C)**2)
        b0 = M_C/z*(s + qn%k)*a0
      else
        if (qn%k < 0) then
          s = -qn%k
          a0 = M_ONE
          b0 = M_ZERO
        else
          s = qn%k
          a0 = M_ZERO
          b0 = M_ONE
        end if
      end if
      w = e + M_TWO*M_C**2 - vp0
      a1 = (w*(s + M_ONE - qn%k)*b0 - z/M_C*(e - vp0)*a0)/(M_TWO*s + M_ONE)/M_C
      b1 = -((e - vp0)*a0 + z*a1)/(s + M_ONE - qn%k)/M_C
      a2 = (w*(s + M_TWO - qn%k)*b1 - z/M_C*(e - vp0)*a1)/(M_TWO*s + M_TWO)/(M_TWO*M_C)
      b2 = -((e - vp0)*a1 + z*a2)/(s + M_TWO - qn%k)/M_C
      a3 = M_ZERO
    case (SCALAR_REL)
      w = e + M_TWO*M_C**2 - vp0
      if (z /= M_ZERO) then
        s = sqrt(qn%l*(qn%l + M_ONE) + M_ONE - (z/M_C)**2)
        a0 = M_ONE
        b0 = M_C/z*(s - M_ONE)*a0
        a1 = (w*b0 - z/M_C*M_TWO*(e + M_C - vp0)*a0)/(M_TWO*s + M_ONE)/M_C
        b1 = M_C/z*(s*a1 - w*b0/M_C)
        a2 = (w*(vp0 - e)*a0/M_C + w*b1 - z/M_C*M_TWO*(e + M_C - vp0)*a1)/(M_FOUR*s + M_FOUR)/M_C
        b2 = M_C/z*((s + M_ONE)*a2 - w*b1/M_C)
        a3 = M_ZERO
      else
        if (qn%l == 0) then
          s = M_ONE
          a0 = M_ONE
          b1 = (vp0 - e)*a0/M_THREE
          a2 = w/(M_TWO*M_C)*b1
          b0 = M_ZERO ; a1 = M_ZERO ; b2 = M_ZERO ; a3 = M_ZERO
        else
          s = qn%l
          b0 = M_ONE
          a1 = w/(s*M_C)*b0
          a0 = M_ZERO ; b1 = M_ZERO ; a2 = M_ZERO ; b2 = M_ZERO ; a3 = M_ZERO
        end if
      end if
    end select
    g(1) = r0**(s - M_ONE)*(a0 + a1*r0 + a2*r0**2 + a3*r0**3)
    f(1) = r0**(s - M_ONE)*(b0 + b1*r0 + b2*r0**2)

    !Set the parameters needed to compute the derivatives of the functions
    call set_derivs_params(qn, e, wave_eq, potential)

    !Integrate the equation from r0 to ri
    ierr = ode_solve(ri, integrator%stp, integrator%evl, integrator%ctrl, (r0+ri)/M_TWO, &
                     integrator%hmin, integrator%nstepmax, nstep, r, g, f, derivs)
    if (ierr /= 0) then
      if (in_debug_mode) then
        call wave_equations_integrator_debug(integrator, r0, ri, nstep, r, g, f)
        call wave_equations_derivs_debug()
      end if
      message(1) = "Error in subtoutine outward_integration. Error message:"
      select case (ierr)
      case (100)
        message(2) = "stepsize underflow"
      case (101)
        message(2) = "integration diverged"
      case default
        call gsl_strerror(ierr, message(2))
      end select
      call write_fatal(2)
    end if

    !Unset the derivatives parameters
    call unset_derivs_params()

    !Copy the functions to new arrays
    allocate(r_out(nstep), g_out(nstep), f_out(nstep))
    r_out = r(1:nstep)
    g_out = g(1:nstep)
    f_out = f(1:nstep)

    !Deallocate arrays
    deallocate(r, f, g)

    call pop_sub()
  end subroutine outward_integration

  subroutine inward_integration(qn, e, wave_eq, potential, integrator, nstep, r_in, g_in, f_in)
    !-----------------------------------------------------------------------!
    ! Integrates the radial wave-equation inward from a point at infinity   !
    ! to the classical turning point.                                       !
    !                                                                       !
    !  qn         - set of quantum numbers                                  !
    !  e          - energy                                                  !
    !  wave_eq    - wave-equation to integrate                              !
    !  potential  - potential object                                        !
    !  integrator - integrator object                                       !
    !  nstep      - number of steps taken by the solver                     !
    !  r_in       - mesh used by the ODE solver                             !
    !  g_in       - wavefunction                                            !
    !  f_in       - wavefunction                                            !
    !-----------------------------------------------------------------------!
    type(qn_type),         intent(in)    :: qn
    real(R8),              intent(in)    :: e
    integer,               intent(in)    :: wave_eq
    type(potential_type),  intent(in)    :: potential
    type(integrator_type), intent(inout) :: integrator
    integer,               intent(out)   :: nstep
    real(R8),              pointer       :: r_in(:), g_in(:), f_in(:)

    integer  :: lpp, ierr
    real(R8) :: vinf, kpp, rinf, ri, minf
    real(R8), allocatable :: r(:), g(:), f(:)

    call push_sub("inward_integration")

    !Allocate work arrays
    allocate(r(integrator%nstepmax), g(integrator%nstepmax), f(integrator%nstepmax))

    !Get the initial and final points
    ri = classical_turning_point(potential, e, qn)
    rinf = practical_infinity(e, wave_eq, integrator%tol)

    !Values of the wavefunction and its derivative at infinity
    r(1) = rinf
    select case (wave_eq)
    case (SCHRODINGER)
      g(1) = SCHRODINGER_FINF
      f(1) = -sqrt(-M_TWO*e)*g(1)
    case (DIRAC)
      !Values of the wavefunctions at infinity
      lpp = int(qn%j + M_HALF)*(qn%j - M_HALF)
      if (lpp /= 0) lpp = lpp/qn%l
      kpp = sqrt(-e)*sqrt(M_TWO + e/M_C2)
      vinf = v(potential, rinf, qn)
      g(1) = DIRAC_FINF

      f(1) = (e - vinf)/M_C* &
           (gsl_sf_bessel_knu_scaled(kpp*rinf, lpp + M_HALF) + (real(qn%k - lpp, R8) - M_ONE)/rinf)/ &
           (kpp*gsl_sf_bessel_knu_scaled(kpp*rinf, lpp + 1.5_r8))*g(1)
    case (SCALAR_REL)
      g(1) = SCHRODINGER_FINF
      vinf = v(potential, rinf, qn)
      minf = M_ONE + (e - vinf)/M_TWO/M_C2
      f(1) = -sqrt(-M_TWO*minf*e)/(M_TWO*minf*M_C)*g(1)
    end select

    !Set the parameters needed to compute the derivatives of the functions
    call set_derivs_params(qn, e, wave_eq, potential)

    !Integrate the equation from rinf to ri
    ierr = ode_solve(ri, integrator%stp, integrator%evl, integrator%ctrl, (ri+rinf)/M_TWO, &
                     integrator%hmin, integrator%nstepmax, nstep, r, g, f, derivs)
    if (ierr /= 0) then
      if (in_debug_mode) then
        call wave_equations_integrator_debug(integrator, rinf, ri, nstep, r, g, f)
        call wave_equations_derivs_debug()
      end if
      message(1) = "Error in subtoutine inward_integration. Error message:"
      select case (ierr)
      case (100)
        message(2) = "stepsize underflow"
      case (101)
        message(2) = "integration diverged"
      case default
        call gsl_strerror(ierr, message(2))
      end select
      call write_fatal(2)
    end if

    !Unset the derivatives parameters
    call unset_derivs_params()

    !Copy the functions to new arrays
    allocate(r_in(nstep), g_in(nstep), f_in(nstep))
    r_in = r(1:nstep)
    g_in = g(1:nstep)
    f_in = f(1:nstep)

    !Deallocate arrays
    deallocate(r, g, f)

    call pop_sub()
  end subroutine inward_integration

  function practical_infinity(e, wave_eq, tol)
    !-----------------------------------------------------------------------!
    ! Returns the value of the practical infinity. The practical infinity   !
    ! is such that wavefunction at the practical infinity is of the order   !
    ! of magnitude of SCHRODINGER_FINF or DIRAC_FINF.                       !
    !                                                                       !
    !  e       - energy                                                     !
    !  wave_eq - wave-equation to use                                       !
    !  tol     - tolerance                                                  !
    !-----------------------------------------------------------------------!
    real(R8), intent(in) :: e
    integer,  intent(in) :: wave_eq
    real(R8), intent(in) :: tol
    real(R8) :: practical_infinity

    real(R8) :: k, x, f, xm, d

    call push_sub("practical_infinity")

    select case (wave_eq)
    case (SCHRODINGER)
      if (e == M_ZERO) then
        practical_infinity = -log(SCHRODINGER_FINF)/sqrt(M_TWO*1.0e-12)
      else
        practical_infinity = -log(SCHRODINGER_FINF)/sqrt(-M_TWO*e)
      end if
    case (DIRAC)
      k = sqrt(-e)*sqrt(M_TWO + e/M_C2)
      x = M_ZERO
      f = -M_HALF*log(x) - k*x - log(DIRAC_FINF)
      d = -log(DIRAC_FINF)/k - x
      do
        d = d*M_HALF
        xm = x + d
        if (f*(-M_HALF*log(xm) - k*xm - log(DIRAC_FINF)) > M_ZERO) x = xm
        if (-M_HALF*log(xm) - k*xm - log(DIRAC_FINF) == M_ZERO .or. d <= tol) exit
      end do
      practical_infinity = xm
    case (SCALAR_REL)
      if (e == M_ZERO) then
        practical_infinity = -log(SCHRODINGER_FINF)/sqrt(M_TWO*(M_ONE - 1.0e-12/M_TWO/M_C**2)*1.0e-12)
      else
        practical_infinity = -log(SCHRODINGER_FINF)/sqrt(-M_TWO*(M_ONE + e/M_TWO/M_C**2)*e)
      end if

    end select

    call pop_sub()
  end function practical_infinity

  subroutine wave_equations_integrator_debug(integrator, ri, rf, nstep, r, g, f)
    !-----------------------------------------------------------------------!
    ! Prints debug information to the "debug_info/integrator" file.         !
    !                                                                       !
    !  integrator - integrator object                                       !
    !  ri         - initial integration point                               !
    !  rf         - final integration point                                 !
    !  nstep      - number of steps taken by the solver                     !
    !  r          - mesh used by the ODE solver                             !
    !  g          - wavefunction                                            !
    !  f          - wavefunction                                            !
    !-----------------------------------------------------------------------!
    type(integrator_type), intent(in) :: integrator
    real(R8),              intent(in) :: ri, rf
    integer,               intent(in) :: nstep
    real(R8),              intent(in) :: r(integrator%nstepmax), g(integrator%nstepmax), f(integrator%nstepmax)

    integer :: unit, iostat, i

    call push_sub("wave_equations_integrator_debug")

    call io_assign(unit)
    open(unit, file='debug_info/integrator', status='unknown', iostat=iostat)
    if (iostat > 0) then
      message(1) = "Unable to open file ''debug_info/integrator''"
      call write_fatal(1)
    end if
    write(unit,'("ODE Integrator tolerance: ",ES10.3E2)') integrator%tol
    write(unit,'("ODE Integrator maximum number of steps: ",I6)') integrator%nstepmax
    write(unit,'("ODE Integrator minimum step size: ",ES10.3E2)') integrator%hmin
    close(unit)

    open(unit, file='debug_info/wavefunctions', status='unknown', iostat=iostat)
    if (iostat > 0) then
      message(1) = "Unable to open file ''debug_info/wavefunctions''"
      call write_fatal(1)
    end if
    write(unit,'("# Integration Starting Point: ",ES10.3E2)') ri
    write(unit,'("# Integration Ending Point:   ",ES10.3E2)') rf
    do i = 1, nstep
      write(unit,'(ES10.3E2,1X,ES10.3E2,1X,ES10.3E2)') r(i), g(i), f(i)
    end do
    close(unit)

    call pop_sub()
  end subroutine wave_equations_integrator_debug

  subroutine wave_equations_integrator_end(integrator_sp, integrator_dp)
    !-----------------------------------------------------------------------!
    ! Frees all the memory associated with the GSL objects used by the      !
    ! ODE solver.                                                           !
    !-----------------------------------------------------------------------!
    type(integrator_type), intent(inout) :: integrator_sp, integrator_dp

    call push_sub("wave_equations_integrator_end")

    !Free GSL objects for simple precision integrations
    call gsl_odeiv_step_free(integrator_sp%stp)
    call gsl_odeiv_evolve_free(integrator_sp%evl)
    call gsl_odeiv_control_free(integrator_sp%ctrl)

    !Free GSL objects for double precision integrations
    call gsl_odeiv_step_free(integrator_dp%stp)
    call gsl_odeiv_evolve_free(integrator_dp%evl)
    call gsl_odeiv_control_free(integrator_dp%ctrl)

    call pop_sub()
  end subroutine wave_equations_integrator_end

end module wave_equations_integrator
